"""Select2 field implementation module."""

from django.forms import ChoiceField


class Select2ListChoiceField(ChoiceField):
    """Allows a list of values to be used with a ChoiceField.

    Avoids unusual things that can happen if Select2ListView is used for
    a form where the text and value for choices are not the same.
    """

    def __init__(self, choice_list=[], required=True, widget=None, label=None,
                 initial=None, help_text='', *args, **kwargs):
        """Use a list to generate choices in a ChoiceField.

        .. py:param choice_list: The list to use to generate choices or a
        function that returns a list.
        """
        if callable(choice_list):
            choices = (
                lambda: [(choice, choice) for choice in choice_list()])
        else:
            choices = [(choice, choice) for choice in choice_list]

        super(Select2ListChoiceField, self).__init__(
            choices=choices, required=required, widget=widget, label=label,
            initial=initial, help_text=help_text, *args, **kwargs
        )


class Select2ListCreateChoiceField(Select2ListChoiceField):
    """Skips validation of choices so any value can be used."""

    def validate(self, value):
        """Do not validate choices but check for empty."""
        super(ChoiceField, self).validate(value)



"""Select2 view implementation."""

import json

from dal.views import BaseQuerySetView, ViewMixin

from django import http
from django.core.exceptions import ImproperlyConfigured
from django.utils.translation import ugettext as _
from django.views.generic.list import View


class Select2ViewMixin(object):
    """View mixin to render a JSON response for Select2."""

    def get_results(self, context):
        """Return data for the 'results' key of the response."""
        return [
            {
                'id': self.get_result_value(result),
                'text': self.get_result_label(result),
            } for result in context['object_list']
        ]

    def get_create_option(self, context, q):
        """Form the correct create_option to append to results."""
        create_option = []
        display_create_option = False
        if self.create_field and q:
            page_obj = context.get('page_obj', None)
            if page_obj is None or page_obj.number == 1:
                display_create_option = True

        if display_create_option and self.has_add_permission(self.request):
            create_option = [{
                'id': q,
                'text': _('Create "%(new_value)s"') % {'new_value': q},
                'create_id': True,
            }]
        return create_option

    def render_to_response(self, context):
        """Return a JSON response in Select2 format."""
        q = self.request.GET.get('q', None)

        create_option = self.get_create_option(context, q)

        return http.HttpResponse(
            json.dumps({
                'results': self.get_results(context) + create_option,
                'pagination': {
                    'more': self.has_more(context)
                }
            }),
            content_type='application/json',
        )


class Select2QuerySetView(Select2ViewMixin, BaseQuerySetView):
    """List options for a Select2 widget."""


class Select2ListView(ViewMixin, View):
    """Autocomplete from a list of items rather than a QuerySet."""

    def get_list(self):
        """"Return the list strings from which to autocomplete."""
        return []

    def get(self, request, *args, **kwargs):
        """"Return option list json response."""
        results = self.get_list()
        create_option = []
        if self.q:
            results = [x for x in results if self.q in x]
            if hasattr(self, 'create'):
                create_option = [{
                    'id': self.q,
                    'text': 'Create "%s"' % self.q,
                    'create_id': True
                }]
        return http.HttpResponse(json.dumps({
            'results': [dict(id=x, text=x) for x in results] + create_option
        }))

    def post(self, request):
        """"Add an option to the autocomplete list.

        If 'text' is not defined in POST or self.create(text) fails, raises
        bad request. Raises ImproperlyConfigured if self.create if not defined.
        """
        if not hasattr(self, 'create'):
            raise ImproperlyConfigured('Missing "create()"')

        text = request.POST.get('text', None)

        if text is None:
            return http.HttpResponseBadRequest()

        text = self.create(text)

        if text is None:
            return http.HttpResponseBadRequest()

        return http.HttpResponse(json.dumps({
            'id': text,
            'text': text,
        }))



"""Widgets for Select2 and django-taggit."""

from dal_select2.widgets import TagSelect2

from django import VERSION


class TaggingSelect2(TagSelect2):
    """Select2 tag widget for tagging's TagField."""

    def render_options(self, *args):
        """Render only selected tags."""
        selected_choices_arg = 1 if VERSION < (1, 10) else 0

        selected_choices = args[selected_choices_arg]

        if selected_choices:
            # Here, selected_choices is a string of comma-separated tags:
            # that's how the tagging field works, otherwise it'd be an empty
            # list because that's how the select field uses for None values
            selected_choices = selected_choices.split(',')

        options = [
            '<option value="%s" selected="selected">%s</option>' % (c, c)
            for c in selected_choices
        ]

        return '\n'.join(options)



"""Checks for the dal_select2 module."""
import os

from django.core import checks


def select2_submodule_check(app_configs, **kwargs):
    """Return an error if select2 is missing."""
    errors = []
    dal_select2_path = os.path.dirname(__file__)
    select2 = os.path.join(
        os.path.abspath(dal_select2_path),
        'static/autocomplete_light/vendor/select2/dist/js/select2.min.js'
    )

    if not os.path.exists(select2):
        errors.append(
            checks.Error(
                'Select2 static files not checked out',
                hint='Run git submodule update --init in DAL ({})'.format(
                    os.path.dirname(dal_select2_path)),
                id='dal_select2.E001',
            )
        )

    return errors



"""Default AppConfig for dal_select2."""
from django.apps import AppConfig
from django.core import checks

from .checks import select2_submodule_check


class DefaultApp(AppConfig):
    """Default app for dal_select2."""

    name = 'dal_select2'

    def ready(self):
        """Register select2_submodule_check."""
        checks.register(select2_submodule_check)



"""Select2 widget implementation module."""

from dal.widgets import (
    QuerySetSelectMixin,
    Select,
    SelectMultiple,
    WidgetMixin
)

from django import forms
from django.utils import six


class Select2WidgetMixin(object):
    """Mixin for Select2 widgets."""

    class Media:
        """Automatically include static files for the admin."""

        css = {
            'all': (
                'autocomplete_light/vendor/select2/dist/css/select2.css',
                'autocomplete_light/select2.css',
            )
        }
        js = (
            'autocomplete_light/jquery.init.js',
            'autocomplete_light/autocomplete.init.js',
            'autocomplete_light/vendor/select2/dist/js/select2.full.js',
            'autocomplete_light/forward.js',
            'autocomplete_light/select2.js',
        )

    autocomplete_function = 'select2'


class Select2(Select2WidgetMixin, Select):
    """Select2 widget for regular choices."""


class Select2Multiple(Select2WidgetMixin, SelectMultiple):
    """Select2Multiple widget for regular choices."""


class ListSelect2(WidgetMixin, Select2WidgetMixin, forms.Select):
    """Select widget for regular choices and Select2."""


class ModelSelect2(QuerySetSelectMixin,
                   Select2WidgetMixin,
                   forms.Select):
    """Select widget for QuerySet choices and Select2."""


class ModelSelect2Multiple(QuerySetSelectMixin,
                           Select2WidgetMixin,
                           forms.SelectMultiple):
    """SelectMultiple widget for QuerySet choices and Select2."""


class TagSelect2(WidgetMixin,
                 Select2WidgetMixin,
                 forms.SelectMultiple):
    """Select2 in tag mode."""

    def build_attrs(self, *args, **kwargs):
        """Automatically set data-tags=1."""
        attrs = super(TagSelect2, self).build_attrs(*args, **kwargs)
        attrs.setdefault('data-tags', 1)
        return attrs

    def value_from_datadict(self, data, files, name):
        """Return a comma-separated list of options.

        This is needed because Select2 uses a multiple select even in tag mode,
        and the model field expects a comma-separated list of tags.
        """
        values = super(TagSelect2, self).value_from_datadict(data, files, name)
        return six.text_type(',').join(values)



"""View for a Select2 widget and QuerySetSequence-based business logic."""

from dal_queryset_sequence.views import BaseQuerySetSequenceView

from dal_select2.views import Select2ViewMixin

from django.template.defaultfilters import capfirst
from django.utils import six


class Select2QuerySetSequenceView(BaseQuerySetSequenceView, Select2ViewMixin):
    """
    Combines support QuerySetSequence and Select2 in a single view.

    Example usage::

        url(
            '^your-generic-autocomplete/$',
            autocomplete.Select2QuerySetSequenceView.as_view(
                queryset=autocomplete.QuerySetSequence(
                    Group.objects.all(),
                    TestModel.objects.all(),
                )
            ),
            name='your-generic-autocomplete',
        )

    It is compatible with the :py:mod:`~dal_select2_queryset_sequence.widgets`
    and the fields of :py:mod:`dal_contenttypes`, suits generic relation
    autocompletes.
    """

    def get_results(self, context):
        """
        Return a list of results usable by Select2.

        It will render as a list of one <optgroup> per different content type
        containing a list of one <option> per model.
        """
        groups = {}

        for result in context['object_list']:
            groups.setdefault(type(result), [])
            groups[type(result)].append(result)

        return [{
            'id': None,
            'text': capfirst(model._meta.verbose_name),
            'children': [{
                'id': self.get_result_value(result),
                'text': six.text_type(result),
            } for result in results]
        } for model, results in groups.items()]



"""Classes for class-based forward declaration."""


class Forward(object):
    """Base class for autocomplete forward declaration."""

    def type(self):
        """Forward type. Should be implemented in subclasses."""
        raise NotImplementedError("Please use one of my subclasses")

    def to_dict(self):
        """Convert to dictionary which will be rendered as JSON."""
        return {
            "type": self.type
        }


class Field(Forward):
    """Forward field value.

    .. py:attribute:: src

        The name of the form field whose value will be forwarded to a view.

    .. py:attribute:: dst

        The name of the key of the forwarded value from the src field in the
        forwarded dictionary. If this value is ``None``, then the key is
        ``src``.
    """

    type = "field"

    def __init__(self, src, dst=None):
        """Instantiate a forwarded field value."""
        self.src = src
        self.dst = dst

    def to_dict(self):
        """Convert to dictionary which will be rendered as JSON."""
        d = super(Field, self).to_dict()

        d.update(src=self.src)
        if self.dst is not None:
            d.update(dst=self.dst)

        return d


class Const(Forward):
    """Forward arbitrary constant value.

    .. py:attribute:: val

        The value to forward. Must be JSON-serializable.

    .. py:attribute:: dst

        The name of the key of the forwarded value.
        ``src``.
    """

    type = "const"

    def __init__(self, val, dst):
        """Instantiate a forwarded constant value."""
        self.val = val
        self.dst = dst

    def to_dict(self):
        """Convert to dictionary which will be rendered as JSON."""
        d = super(Const, self).to_dict()

        d.update(val=self.val)
        d.update(dst=self.dst)

        return d



"""Widgets for Select2 and django-taggit."""

from dal_select2.widgets import TagSelect2

from django import VERSION
from django.utils import six


class TaggitSelect2(TagSelect2):
    """Select2 tag widget for taggit's TagField."""

    def render_options(self, *args):
        """Render only selected tags."""
        selected_choices_arg = 1 if VERSION < (1, 10) else 0
        selected_choices = args[selected_choices_arg]

        # When the data hasn't validated, we get the raw input here
        if isinstance(selected_choices, six.text_type):
            choices = [c.strip() for c in selected_choices.split(',')]
        else:
            # Filter out None values, not needed for autocomplete
            choices = [c.tag.name for c in selected_choices if c]

        options = [
            '<option value="%s" selected="selected">%s</option>' % (
                c, c) for c in choices
        ]

        return '\n'.join(options)



"""Test case for autocomplete implementations."""

import os
import uuid

from django import VERSION
from django.contrib.contenttypes.models import ContentType
from django.contrib.staticfiles.testing import StaticLiveServerTestCase
from django.core.urlresolvers import reverse
from django.utils import six

from splinter import Browser


GLOBAL_BROWSER = None


class AutocompleteTestCase(StaticLiveServerTestCase):
    """Provide a class-persistent selenium instance and assertions."""

    @classmethod
    def setUpClass(cls):
        """Instanciate a browser for the whole test session."""
        global GLOBAL_BROWSER

        if GLOBAL_BROWSER is None:
            GLOBAL_BROWSER = Browser(os.environ.get('BROWSER', 'firefox'))
        cls.browser = GLOBAL_BROWSER

        super(AutocompleteTestCase, cls).setUpClass()

    def get(self, url):
        """Open a URL."""
        self.browser.visit('%s%s' % (
            self.live_server_url,
            url
        ))

        if '/admin/login/' in self.browser.url:
            # Should be pre-filled by HTML template
            # self.browser.fill('username', 'test')
            # self.browser.fill('password', 'test')
            self.browser.find_by_value('Log in').click()

    def click(self, selector):
        """Click an element by css selector."""
        self.browser.find_by_css(selector).click()

    def enter_text(self, selector, text):
        """Enter text in an element by css selector."""
        self.browser.find_by_css(selector).type(text)

    def assert_not_visible(self, selector):
        """Assert an element is not visible by css selector."""
        e = self.browser.find_by_css(selector)
        assert not e or e.visible is False

    def assert_visible(self, selector):
        """Assert an element is visible by css selector."""
        assert self.browser.find_by_css(selector).visible is True


class AdminMixin(object):
    """Mixin for tests that should happen in ModelAdmin."""

    def get_modeladmin_url(self, action, **kwargs):
        """Return a modeladmin url for a model and action."""
        return reverse('admin:%s_%s_%s' % (
            self.model._meta.app_label,
            self.model._meta.model_name,
            action
        ), kwargs=kwargs)

    def fill_name(self):
        """Fill in the name input."""
        i = self.id()
        half = int(len(i))
        not_id = i[half:] + i[:half]
        self.browser.fill('name', not_id)


class OptionMixin(object):
    """Mixin to make a unique option per test."""

    def create_option(self):
        """Create a unique option from self.model into self.option."""
        unique_name = six.text_type(uuid.uuid1())

        if VERSION < (1, 10):
            # Support for the name to be changed through a popup in the admin.
            unique_name = unique_name.replace('-', '')

        option, created = self.model.objects.get_or_create(
            name=unique_name)
        return option


class ContentTypeOptionMixin(OptionMixin):
    """Same as option mixin, with content type."""

    def create_option(self):
        """Return option, content type."""
        option = super(ContentTypeOptionMixin, self).create_option()
        ctype = ContentType.objects.get_for_model(option)
        return option, ctype



"""
Widget mixin that only renders selected options with QuerySetSequence.

For details about why this is required, see :mod:`dal.widgets`.
"""

from dal.widgets import WidgetMixin

from django import forms
from django.contrib.contenttypes.models import ContentType
from django.utils import six


class QuerySetSequenceSelectMixin(WidgetMixin):
    """Support QuerySetSequence in WidgetMixin."""

    def filter_choices_to_render(self, selected_choices):
        """Overwrite self.choices to exclude unselected values."""
        if len(selected_choices) == 1 and not selected_choices[0]:
            selected_choices = []

        ctype_models = {}

        for choice in selected_choices:
            ctype_pk, model_pk = choice.split('-')
            ctype_pk = int(ctype_pk)
            ctype_models.setdefault(ctype_pk, [])
            ctype_models[ctype_pk].append(model_pk)

        self.choices = []
        ctype = ContentType.objects.get_for_id
        for ctype_pk, ids in ctype_models.items():
            results = ctype(ctype_pk).model_class().objects.filter(pk__in=ids)

            self.choices += [
                ('%s-%s' % (ctype_pk, r.pk), six.text_type(r))
                for r in results
            ]


class QuerySetSequenceSelect(QuerySetSequenceSelectMixin,
                             forms.Select):
    """Select widget for QuerySetSequence choices."""


class QuerySetSequenceSelectMultiple(QuerySetSequenceSelectMixin,
                                     forms.SelectMultiple):
        """SelectMultiple widget for QuerySetSequence choices."""



"""Base views for autocomplete widgets."""

import json

from django import http
from django.contrib.auth import get_permission_codename
from django.core.exceptions import ImproperlyConfigured
from django.utils import six
from django.views.generic.list import BaseListView


class ViewMixin(object):
    """Common methods for autocomplete views.

    .. py:attribute:: forwarded

        Dict of field values that were forwarded from the form, may be used to
        filter autocompletion results based on the form state. See
        ``linked_data`` example for reference.

    .. py:attribute:: q

        Query string as typed by the user in the autocomplete field.
    """

    def dispatch(self, request, *args, **kwargs):
        """Set :py:attr:`forwarded` and :py:attr:`q`."""
        self.forwarded = json.loads(request.GET.get('forward', '{}'))
        if request.method == 'POST':
            self.forwarded = json.loads(request.POST.get('forward', '{}'))
        self.q = request.GET.get('q', '')
        return super(ViewMixin, self).dispatch(request, *args, **kwargs)


class BaseQuerySetView(ViewMixin, BaseListView):
    """Base view to get results from a QuerySet.

    .. py:attribute:: create_field

        Name of the field to use to create missing values. For example, if
        create_field='title', and the user types in "foo", then the
        autocomplete view will propose an option 'Create "foo"' if it can't
        find any value matching "foo". When the user does click 'Create "foo"',
        the autocomplete script should POST to this view to create the object
        and get back the newly created object id.
    """

    paginate_by = 10
    context_object_name = 'results'
    create_field = None

    def has_more(self, context):
        """For widgets that have infinite-scroll feature."""
        return context['page_obj'].has_next()

    def get_result_value(self, result):
        """Return the value of a result."""
        return result.pk

    def get_result_label(self, result):
        """Return the label of a result."""
        return six.text_type(result)

    def get_queryset(self):
        """Filter the queryset with GET['q']."""
        qs = super(BaseQuerySetView, self).get_queryset()

        if self.q:
            qs = qs.filter(name__icontains=self.q)

        return qs

    def create_object(self, text):
        """Create an object given a text."""
        return self.get_queryset().create(**{self.create_field: text})

    def has_add_permission(self, request):
        """Return True if the user has the permission to add a model."""
        if not request.user.is_authenticated():
            return False

        opts = self.get_queryset().model._meta
        codename = get_permission_codename('add', opts)
        return request.user.has_perm("%s.%s" % (opts.app_label, codename))

    def post(self, request):
        """Create an object given a text after checking permissions."""
        if not self.has_add_permission(request):
            return http.HttpResponseForbidden()

        if not self.create_field:
            raise ImproperlyConfigured('Missing "create_field"')

        text = request.POST.get('text', None)

        if text is None:
            return http.HttpResponseBadRequest()

        result = self.create_object(text)

        return http.JsonResponse({
            'id': result.pk,
            'text': six.text_type(result),
        })
